#include <alloc.h>
#include <stdio.h>
#include <stdlib.h>
#include "struct.hpp"
#include "motor_i.hpp"

void Ventana_errores(char * Linea1,char *Linea2,char *Linea3); //botones.cpp


/* 
	FUNCION : pertenencia;
	OBJETIVO : Dato un conjunto fuzzy y un valor debe hallar el valor de la 
		funcin de asociacin del valor al conjunto y devolverla
	PARAMETROS:
		cnj   --> conjunto fuzzy
		value --> valor determinado de la  vble en un determinado momento;
*/

double pertenencia(cnj_fuzzy cnj,double value)
	{
	 double BaseL, BaseR, RatioL = 0, RatioR = 0;

	 if ( (value < cnj.left) || ( value > cnj.right))
		return 0;  /* Si el valor esta fuera de los lmites del cnj */

	 BaseL = cnj.top_left - cnj.left;
	 BaseR = cnj.right - cnj.top_right;
	 if ( BaseL > 0 ) 
		RatioL = ( 100 / BaseL);
	 if ( BaseR > 0 ) 
		RatioR = ( 100 / BaseR); /* se calcula la pentiente de los borde de los cnj*/

	 if ( cnj.tipo == TRIANGULAR )
		{
		 if ( value < cnj.top_left)
			return (( value - cnj.left) * RatioL);
		 else
			if ( value > cnj.top_left)
				return (( cnj.right - value) * RatioR);
			else
				return 100;
		}
	 else  // trapeziodal
		{
		 if ( (value >= cnj.top_left ) && ( value <= cnj.top_right))
			return (100);
		 if ( value <= cnj.top_left)
			return (( value - cnj.left) * RatioL);
		 else
			return (( cnj.right - value) * RatioR);

		}
	}




/*
	FUNCION : valor_entrada
	OBJETIVO : Funcin que dada una variable y un valor determinado, halla la 
		funcin de pertenencia a todos sus cnj.
	FUNCIONES USADAS:
		pertenencia.
	PARAMETROS:
		vble  --> variable fuzzy
		value --> valor determinado de la  vble en un determinado momento;
*/

void valor_entrada(vble v, double value )
	{
	 cnj_fuzzy *p;

	 p = v.conjuntos;
	 while ( p != NULL)
		{
		 p->grado_pertenencia = pertenencia(*p,value);
		 p = p->next;
		}
	}



double minimo(double a, double b)
	{
	 return ( (a < b) ? a : b);
	}

double maximo(double a, double b)
	{
	 return ( (a > b) ? a : b);
	}

/* 
	FUNCION : calculo_regla
	OBJETIVO : Realizar el cmputo de una regla. Para ello se recorre la 
		estructura de rbol donde se almacena la regla realizando en cada
		caso la operacin que se indique en el nodo.
		El valor que resulta se almacenara el en cnj correspondiente de 
		la variable de salida
	PARAMETROS:
		r --> regla a evaluar;
*/

double calculo_regla(regla r)
	{
	 regla *p;
	 double result;
	 cnj_fuzzy *cnj;
	 lista_per *lp;
	 lista_per *q;
	 
	 switch (r.operador)
		{
		 case EQ:
					p = r.nodo.nohoja.der; /* La parte derache tendra la variable de salida */	
					/* El grado de pertenencia de la vble de salida ser la 
						evaluacin del lado izquierdo del arbol */
					result = calculo_regla(*r.nodo.nohoja.izq);
					lp = (lista_per *)farmalloc(sizeof(lista_per));
					if ( lp == NULL)
						{
						 printf("\nERROR: Memoria insuficiente");
						 exit(-1);
						}
					lp->grado = result;
					lp->next = NULL;
					cnj = p->nodo.hoja.cnj;
					if ( cnj->ptr_lista == NULL)
						{
						 cnj->ptr_lista = lp;
						}
					else
						{
						 q = cnj->ptr_lista;
						 while ( q->next != NULL)
							q = q->next;
						 q->next = lp;
						}
					return 0;
		 case AND:
					return ( minimo(calculo_regla(*r.nodo.nohoja.izq),calculo_regla(*r.nodo.nohoja.der)));
		 case OR :
					return ( maximo(calculo_regla(*r.nodo.nohoja.izq),calculo_regla(*r.nodo.nohoja.der)));
		 case NOT:
					return ( (double)100 - calculo_regla(*r.nodo.nohoja.izq));
		 case PER: /* se trata de un nodo hoja y se devuelve el valor de
						  pertenencia de la variable al cnj
					  */
					return ( r.nodo.hoja.cnj->grado_pertenencia);
		}
	 return 0;
	}





/*
	FUNCION : init_cnj
	OBJETIVO : Calcular el area del cnj y el area debajo de la curva
		P(x)*x Para calcular luego media del producto y la suma.
	PARAMETROS:
		cnj --> conjunto a inicializar;
*/
void init_cnj(cnj_fuzzy & cnj)
	{
	 double BaseL, BaseR, BaseC;
	 double temp;
	 BaseL = cnj.top_left - cnj.left;
	 BaseR = cnj.right - cnj.top_right;

	 if ( cnj.tipo == TRIANGULAR)
		{
		 temp = 0;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left - cnj.left * cnj.left) / (double)2;
			 temp -= cnj.left * BaseL;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += BaseR;
			 temp  -= ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / ((double)2 * BaseR);
			 temp += cnj.top_right;
			}
		 cnj.area= (double)100 * temp;

		 temp = 0;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left * cnj.top_left - cnj.left * cnj.left * cnj.left) / (double)3;
			 temp -= (cnj.left * (cnj.top_left * cnj.top_left - cnj.left * cnj.left)) / (double)2;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / (double)2;
			 temp  -= ( cnj.right * cnj.right * cnj.right - cnj.top_right * cnj.top_right * cnj.top_right ) / ((double)3 * BaseR);
			 temp += ( cnj.top_right * ( cnj.right * cnj.right - cnj.top_right * cnj.top_right)) / ( (double)2 *BaseR);
			}
		 cnj.px = (double)100 * temp;
		 /* Responde a la frmula

						 100
				px = ---(BaseL^2 + BaseR^2)
							3
		 */
		}
	 else
		{ /* tipo TRAPEZOIDAL */
		 BaseC = cnj.top_right - cnj.top_left;
		 cnj.area =(( 2 * BaseC + BaseL + BaseR) * (double)100) / (double)2;
		 temp = 0;
		 if ( BaseL != 0)
			{
			 temp  = (cnj.top_left * cnj.top_left * cnj.top_left - cnj.left * cnj.left * cnj.left) / (double)3;
			 temp -= (cnj.left * (cnj.top_left * cnj.top_left - cnj.left * cnj.left)) / (double)2;
			 temp  = temp / BaseL;
			}
		 if ( BaseR != 0)
			{
			 temp += ( cnj.right * cnj.right - cnj.top_right * cnj.top_right ) / (double)2;
			 temp  -= ( cnj.right * cnj.right * cnj.right - cnj.top_right * cnj.top_right * cnj.top_right ) / ((double)3 * BaseR);
			 temp += ( cnj.top_right * ( cnj.right * cnj.right - cnj.top_right * cnj.top_right)) / ( (double)2 *BaseR);
			}
		 if ( BaseC != 0)
			 temp += (cnj.top_right * cnj.top_right - cnj.top_left * cnj.top_left) / (double) 2;
		 cnj.px = (double)100 * temp;
		 /* Responde a la frmula

							 BaseL^2 + BaseR^2        BaseC^2
			 px= 100 ( --------------------- + -------------- )
										3		              2
		 */
		}
	}



/*
	FUNCION : media_producto_suma
	OBJETIVO : Funcion que calcula en valor de una vble de salida aplicando
		el mtodo de la media del producto y la suma.
	PARAMETROS:
		cnj --> conjunto a inicializar;
*/

double media_producto_suma(vble v)
	{
	 extern double valor_defecto;
	 cnj_fuzzy *p;
	 lista_per *q,*r;
	 double denominador = 0;
	 double numerador= 0;
			/* de la exprecin de la media de la suma de los productos */

	 p = v.conjuntos;
	 while ( p != NULL)
		{
		 q = p->ptr_lista;
		 while ( q != NULL)
			{
			 numerador += q->grado * p->px;
			 denominador +=  q->grado * p->area;
			 r = q->next;
			 farfree(q);
			 q = r;
			}
		 p->ptr_lista = NULL;
		 p = p->next;
		}
	 if ( (denominador == 0) || (numerador == 0))
		 return valor_defecto ;
	 return ( numerador / denominador);
	}



int n_entradas ;
int n_salidas  ;

void motor_inferencia(double *vbles_entrada, double *vbles_salida)
   {
	 extern vble *vble_entrada;
	 extern vble *vble_salida;
	 extern lis_reg *lista_reglas;
	 extern int n_entradas ;
	 extern int n_salidas  ;
	 int i;
	 vble *p;
    lis_reg *l;


	 // contar variables de entrada

	 p = vble_entrada;
	 for ( i =0; i < n_entradas; i ++, p = p->next )
		{
		 valor_entrada(*p,vbles_entrada[i]);
		}
    l = lista_reglas;
    while(l != NULL)
		{
		 calculo_regla(*(l->reg));
		 l = l->next;
		}
    p = vble_salida;
	 for ( i = 0; i < n_salidas; i ++, p = p->next )
		{
		 vbles_salida[i] = media_producto_suma(*p);
		}
  }



// funcion para leer datos de un fichero de entrada

char init(void)
	{
	 extern unsigned char far *buf1;
	 extern unsigned char far *buf2;
	 extern unsigned char far *buf3;
	 extern unsigned char far *buf4;

	 extern vble *vble_entrada;
	 extern vble *vble_salida;
	 extern lis_reg *lista_reglas;
	 extern int n_entradas ;
	 extern int n_salidas ;
	 int i;
	 vble *p;
	 int fd;
	 lis_reg *l,*anl;
	 int n_reglas = 0;
	 vble *v,*anv;



    // inicializar los conjuntos

	 p = vble_entrada;
	 while ( p != NULL)
		{
		 cnj_fuzzy *c;
		 c = p->conjuntos;
		 while ( c != NULL)
			{
			 init_cnj(*c);
			 c = c->next;
			}
		 p = p->next;
		}

	 p = vble_salida;
	 while ( p != NULL)
		{
		 cnj_fuzzy *c;
		 c = p->conjuntos;
		 while ( c != NULL)
			{
			 init_cnj(*c);
			 c = c->next;
			}
		 p = p->next;
		}
	 p = vble_entrada;
	 n_entradas= 0;
	 while ( p != NULL)
		{
		 n_entradas ++;
		 p = p->next;
		}

	 if ( vble_entrada == 0 )
		{
		 Ventana_errores("No existen variables de", "entrada definidas en","este sistema");
		 return -1;
		}
	 // contar variables de salida

	 p = vble_salida;
	 n_salidas= 0;
	 while ( p != NULL)
		{
		 n_salidas ++;
		 p = p->next;
		}

	 if ( vble_salida == 0 )
		{
		 Ventana_errores("No existen variables de", "salida definidas en","este sistema");
		 return -1;
		}

   return 0;

	}
